@@ -5,7 +5,7 @@
  *	  to avoid conflicts with any uuid_t type that might be defined by
  *	  the system headers.
  *
- * Copyright (c) 2007-2021, PostgreSQL Global Development Group
+ * Copyright (c) 2007-2023, PostgreSQL Global Development Group
  *
  * src/include/utils/uuid.h
  *
@@ -23,9 +23,20 @@ typedef struct pg_uuid_t
 } pg_uuid_t;
 
 /* fmgr interface macros */
-#define UUIDPGetDatum(X)		PointerGetDatum(X)
+static inline Datum
+UUIDPGetDatum(const pg_uuid_t *X)
+{
+	return PointerGetDatum(X);
+}
+
 #define PG_RETURN_UUID_P(X)		return UUIDPGetDatum(X)
-#define DatumGetUUIDP(X)		((pg_uuid_t *) DatumGetPointer(X))
+
+static inline pg_uuid_t *
+DatumGetUUIDP(Datum X)
+{
+	return (pg_uuid_t *) DatumGetPointer(X);
+}
+
 #define PG_GETARG_UUID_P(X)		DatumGetUUIDP(PG_GETARG_DATUM(X))
 
 #endif							/* UUID_H */
